/*-
 * Copyright 2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#include <string.h>

#include "vcl.h"
#include "vas.h"
#include "vdef.h"
#include "vrt.h"
#include "cache/cache.h"
#include "vsb.h"

#include "vcc_if.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod blob error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod blob error: " fmt, __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
        ERR((ctx), msg ", out of space")

#define VERRNOMEM(ctx, fmt, ...) \
        VERR((ctx), fmt ", out of space", __VA_ARGS__)

static const struct vmod_priv null_blob[1] =
{
        {
                .priv = "",
                .len = 0,
                .free = NULL
        }
};

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		VSLv(SLT_VCL_Error, 0, fmt, args);
	va_end(args);

	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		va_start(args, fmt);
		VSB_vprintf(ctx->msg, fmt, args);
		va_end(args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
}

VCL_BOOL
vmod_same(VRT_CTX __attribute__((unused)), VCL_BLOB b1, VCL_BLOB b2)
{
	if (b1 == NULL && b2 == NULL)
		return 1;
	if (b1 == NULL || b2 == NULL)
		return 0;
	return b1->len == b2->len && b1->priv == b2->priv;
}

VCL_BOOL
vmod_equal(VRT_CTX __attribute__((unused)), VCL_BLOB b1, VCL_BLOB b2)
{
	if (b1 == NULL && b2 == NULL)
		return 1;
	if (b1 == NULL || b2 == NULL)
		return 0;
	if (b1->len != b2->len)
		return 0;
	if (b1->priv == b2->priv)
		return 1;
	if (b1->priv == NULL || b2->priv == NULL)
		return 0;
	return (memcmp(b1->priv, b2->priv, b1->len) == 0);
}

VCL_INT
vmod_length(VRT_CTX __attribute__((unused)), VCL_BLOB b)
{
	if (b == NULL)
		return 0;
	return b->len;
}

VCL_INT
vmod_integer(VRT_CTX, VCL_BLOB b, VCL_ENUM padding)
{
	VCL_INT i = 0;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (b == NULL || b->len == 0 || b->priv == NULL) {
		ERR(ctx, "blob is empty in blob.integer()");
		return 0;
	}
	assert(b->len >= 0);

	if ((unsigned)b->len >= sizeof(VCL_INT))
		return *((VCL_INT *) b->priv);
	for (int j = 0; j < b->len; j++)
		i |= ((unsigned char *)b->priv)[j] << ((b->len - j - 1) * 8);
	if (padding[0] == 'R')
		i <<= (sizeof(VCL_INT) - b->len) * 8;
	return i;
}

VCL_BLOB
vmod_subblob(VRT_CTX, VCL_BLOB b, VCL_BYTES n, VCL_BYTES off)
{
	uintptr_t snap;
	struct vmod_priv *sub;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(n >= 0);
	assert(off >= 0);

	if (b == NULL || b->len == 0 || b->priv == NULL) {
		ERR(ctx, "blob is empty in blob.subblob()");
		return NULL;
	}
	assert(b->len >= 0);
	if (off + n > b->len) {
		VERR(ctx, "size %d from offset %d requires more bytes than "
		     "blob length %d in blob.subblob()", n, off, b->len);
		return NULL;
	}

	if (n == 0)
		return null_blob;

	snap = WS_Snapshot(ctx->ws);
	if ((sub = WS_Alloc(ctx->ws, sizeof(*sub))) == NULL) {
		ERRNOMEM(ctx, "Allocating BLOB result in blob.subblob()");
		return NULL;
	}
	if ((sub->priv = WS_Alloc(ctx->ws, n)) == NULL) {
		VERRNOMEM(ctx, "Allocating %d bytes in blob.subblob()", n);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	memcpy(sub->priv, b->priv + off, n);
	sub->len = n;
	return sub;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
